/************************************************************************
    frap/defrap - fractal image compressor/decompressor
    Copyright (C) 1997-2000 Matthias Ruhl

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    The author can be contacted at:
      MIT Laboratory for Computer Science
      200 Technology Square
      Cambridge, MA 02139
      USA
    or via e-mail: ruhl@theory.lcs.mit.edu
*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#ifdef GETOPT
#include <getopt.h>
#endif
#include <math.h>
#ifdef ZLIB
#include <zlib.h>
#endif
#include "defrap.h"
#include "fileio.h"
#include "version.h"

char *input = "frap.out", *output = "frap.ppm";
int iterations = 10, showpart = 0;

pixel ***image,***image2;
int width,height,depth,basesize,rangenum,s_bits,o_bits;
char **kn,**kn2,**know;
rangeT *r;

typedef struct { int x,y; } point;

point stack[10000];

int stacktop;

void sinit()
{
  stacktop = 0;
}

int sempty()
{
  return !stacktop;
}

void spop(int *x, int *y)
{
  if(stacktop) 
    stacktop--;
  else
    fprintf(stderr,"INTERNAL ERROR: Stack underflow!\n");
  *x = stack[stacktop].x;
  *y = stack[stacktop].y;
}

void spush(int x, int y)
{
  stack[stacktop].x = x;
  stack[stacktop].y = y;
  kn2[y][x] |= 16;
  stacktop++;
}

typedef struct 
{
  int n,freq;
} freqT;

int compare1(const void *v1, const void *v2)
{
  freqT *a = (freqT *) v1;
  freqT *b = (freqT *) v2;
  if(a->freq != b->freq) return b->freq - a->freq;
  return a->n - b->n;
}

int compare2(const void *v1, const void *v2)
{
  freqT *a = (freqT *) v1;
  freqT *b = (freqT *) v2;
  return a->n - b->n;
}

void check(int x, int y, int val)
{
  if((val & 1) && !(know[y][x] & 1)) 
    { 
      kn[(y+height-1)%height][x] |= 4;
      kn2[(y+height-1)%height][x] |= 4;
    }
  if((val & 2) && !(know[y][x] & 2)) 
    {
      kn[y][(x+1)%width] |= 8;
      kn2[y][(x+1)%width] |= 8;
    }
  if((val & 4) && !(know[y][x] & 4))
    {    
      kn[(y+1)%height][x] |= 1;
      kn2[(y+1)%height][x] |= 1;
    }
  if((val & 8) && !(know[y][x] & 8))
    {
      kn[y][(x+width-1)%width] |= 2;
      kn2[y][(x+width-1)%width] |= 2;
    }
  know[y][x] = 15;
  know[(y+1)%height][x] |= 1;
  know[y][(x+width-1)%width] |= 2;
  know[(y+height-1)%height][x] |= 4;
  know[y][(x+1)%width] |= 8;
  kn[y][x] |= val;
}

void read_part(FILE *inp)
{
  int start[100][3],startnum,i,j;
  int perm[2][8],freq[7];
  freqT f[8];
  int code=0,dir,x,y,tmp,b,val,nst2,**sol,codelen,codepos=0,usegnu,zbits;
  char *codebuf = NULL;
  unsigned char *buffer = NULL;
#ifdef ZLIB
  unsigned char *zbuffer;
  long unsigned int buflen,zbuflen;
#endif
  int *pp,ppnum,x0,y0;

  sol = (int **) malloc(height*sizeof(int *));
  sol[0] = (int *) malloc(width*height*sizeof(int));

  /* 1. First compute the frequecies */

  startnum = unpack((int)ceil(log(width*height)/log(2.0)),inp);

  for(i=0;i<startnum;i++)
    {
      start[i][0] = unpack((int)ceil(log(width)/log(2.0)),inp);
      start[i][1] = unpack((int)ceil(log(height)/log(2.0)),inp);
      start[i][2] = unpack(4,inp);
    }
  usegnu = unpack(1,inp);
#ifndef ZLIB
  if(usegnu)
    {
      printf("ERROR: This code cannot be decompressed without the zlib!\n");
      exit(1);
    }
#else
  if(usegnu)
    {
      zbuffer = (unsigned char *) malloc(2*width*height);
      buffer = (unsigned char *) malloc(2*width*height);
      zbuflen = fread(zbuffer,1,width*height,inp);
      buflen = width*height;
      uncompress(buffer,&buflen,zbuffer,zbuflen);
      codelen = buflen;
      free(zbuffer);
    }
#endif
  else
    {
      for(i=0;i<8;i++) f[i].n = i;
      codelen = f[0].freq = 0;
      for(i=0;i<7;i++) 
	codelen +=
	  f[i+1].freq = freq[i] = 
	  unpack((int)ceil(log(width*height)/log(2.0)),inp);
      codebuf = (char *) malloc(codelen*sizeof(char));
  
      for(i=0;i<8;i++) perm[0][i] = perm[1][i] = -1;
      /* pass 1 */
      qsort(f,8,sizeof(freqT),compare1);
      qsort(f,4,sizeof(freqT),compare2);
      for(i=0;i<4;i++)
	perm[0][f[i].n] = i;
      /* pass 2 */
      qsort(f,4,sizeof(freqT),compare1);
      qsort(f,2,sizeof(freqT),compare2);
      for(i=0;i<2;i++)
	perm[1][f[i].n] = i;
    }

  /* 2. Allocate arrays for decoding */

  kn = (char **) malloc(height*sizeof(char *));
  kn[0] = (char *) calloc(width*height,sizeof(char));
  kn2 = (char **) malloc(height*sizeof(char *));
  kn2[0] = (char *) calloc(width*height,sizeof(char));
  know = (char **) malloc(height*sizeof(char *));
  know[0] = (char *) calloc(width*height,sizeof(char));
  for(i=1;i<height;i++) 
    {
      kn[i] = width+kn[i-1];
      kn2[i] = width+kn2[i-1];
      know[i] = width+know[i-1];
      sol[i] = width+sol[i-1];
    }

  /* 3. Now decode the range boundaries */

  if(!usegnu)
    {
      arith_init(0,inp);
      arith_code(codebuf,codelen,freq);
    }
  codepos = zbits = 0;

  sinit(); x = y = dir = nst2 = 0;
  while(nst2<startnum)
    {
      x = start[nst2][0];
      y = start[nst2][1];
      tmp = start[nst2][2];
      nst2++;
      kn2[y][x] = (char) tmp;
      check(x,y,tmp);
      spush(x,y);
      while(!sempty())
	{
	  while(!sempty()) { spop(&x,&y); if(kn2[y][x] & 15) break; }
	  if((kn2[y][x] & 15) == 0) break;
	  dir = 0;
	  while(1)
	    {
	      while(0 == (kn2[y][x] & 1<<dir)) dir = (dir+1)%4;
	      kn2[y][x] -= 1<<dir; /* do not use that direction again */
	      if(kn2[y][x] & 15) spush(x,y);
	      switch(dir)
		{
		case 0: y = (y+height-1)%height; break;
		case 1: x = (x+1)%width; break;
		case 2: y = (y+1)%height; break;
		case 3: x = (x+width-1)%width; break;
		}
	      if(kn2[y][x] & (1<<((dir+2)%4)))
		kn2[y][x] -= 1<<((dir+2)%4);
	      else fprintf(stderr,"INTERNAL ERROR #1\n");
	      if(kn2[y][x] & 16) break;
	      for(b=i=0;i<4;i++)
		if((know[y][x] & 1<<i) == 0) b++;
	      if(!b) continue;
	      
	      if(!usegnu)
		{
		  code = codebuf[codepos++]+1;
		  if(b!=3) code = perm[2-b][code];
		  if(code == -1) fprintf(stderr,"INTERNAL ERROR #2\n");
		}
	      val = kn[y][x]; b = 0;
	      
	      for(b=i=0;i<4;i++)
		if(!(know[y][x] & 1<<(dir+3+i)%4))
		  {
		    if((usegnu && (buffer[zbits/8] & (1<<(zbits%8)))) ||
		       (!usegnu && (code & 1<<b)))
		      {
			val |= 1<<(dir+3+i)%4;
			kn2[y][x] |= 1<<(dir+3+i)%4;
		      }
		    zbits++;
		    b++;
		  }
	      check(x,y,val);
	      if(kn2[y][x] & (1<<((dir+2)%4)))
		fprintf(stderr,"INTERNAL ERROR #3\n");
	    }
	}
    }
  free(buffer);

  /* 4. Transform range boundaries in block lists */

  i = 0;
  pp = (int *) malloc(width*height*sizeof(int));
  for(x=0;x<width;x++)
    for(y=0;y<height;y++) know[y][x] = 0;
  sinit();
  for(y0=0;y0<height;y0++)
    for(x0=0;x0<width;x0++)
      if(!know[y0][x0])
	{
	  ppnum = 0;
	  spush(x0,y0);
	  while(!sempty())
	    {
	      spop(&x,&y);
	      if(!know[y][x])
		{
		  know[y][x] = 1;
		  sol[y][x] = i;
		  pp[ppnum++] = y*width+x;
		  if((kn[y][x] & 1) == 0) spush((x+1)%width,y);
		  if((kn[y][x] & 8) == 0) spush(x,(y+1)%height);
		  if((kn[y][(x+width-1)%width] & 1) == 0) spush((x+width-1)%width,y);
		  if((kn[(y+height-1)%height][x] & 8) == 0) spush(x,(y+height-1)%height);
		}
	    }
	  r[i].size = ppnum;
	  r[i].pos = (int *) malloc(ppnum*sizeof(int));
	  for(j=0;j<ppnum;j++) r[i].pos[j] = pp[j];
	  i++;
	}
  free(pp);
}

int bound(pixel p)
{
  if(p < 0) return 0;
  if(p > 255) return 255;
  return (int)rint(p);
}

int read_data(char *fname)
{
  FILE *inp;
  int i,qscaling,qoffset,dombits;
#ifdef ZLIB
  long unsigned int buflen,zbuflen;
  unsigned char *zbuffer,*buffer;
#endif
  char id[6];

  inp = fopen(fname,"rb");
  if(!inp) { printf("Could not open file %s!\n",fname); return 0; }

  /* 1. Read in general image information */

  fread(id,1,5,inp); id[5] = 0;
  if(strcmp(id,"FRAP0"))
    { printf("Wrong input file format!\n"); fclose(inp); return 0; }

  packpos  = 0;
  width    = unpack(12,inp);
  height   = unpack(12,inp);
  depth    = unpack(1,inp)?3:1;
  basesize = unpack(8,inp);
  s_bits   = unpack(4,inp);
  o_bits   = unpack(4,inp);
  rangenum = unpack((int)ceil(log(width*height)/log(2.0)),inp);

  r = (rangeT *) malloc(rangenum*sizeof(rangeT));
  if(r == NULL) { printf("Out of memory!\n"); return 0; }

  /* 2. Read in transformation parameters */

  dombits = (int)ceil(log(width*height)/log(2.0)-2.0);

  for(i=0;i<rangenum;i++)
    {
      qscaling    = unpack(s_bits,inp);
      qoffset     = unpack(o_bits,inp);
      r[i].rot    = unpack(3,inp);
      r[i].domain = unpack(dombits,inp);

      r[i].scaling = (double)qscaling/(double)(1<<s_bits)*2.0-1.0;
      r[i].offset = (double)qoffset/(double)((1<<o_bits)-1)*((1.0+fabs(r[i].scaling))*255.0);
      if (r[i].scaling > 0.0) r[i].offset -= r[i].scaling*255.0;
    }

  /* 3. Read in color information, if necessary */

  if(depth == 3)
    {
      if(unpack(1,inp))
	{
#ifdef ZLIB
	  zbuflen = unpack((int)ceil(log(2*width*height)/log(2.0)),inp);
	  zbuffer = (unsigned char *) malloc(zbuflen);
	  buffer = (unsigned char *) malloc(2*width*height);
	  fread(zbuffer,1,zbuflen,inp);
	  packpos = 0;
	  buflen = 2*width*height;
	  uncompress(buffer,&buflen,zbuffer,zbuflen);
	  for(i=0;i<rangenum;i++)
	    {
	      r[i].umean = buffer[2*i] - 127;
	      r[i].vmean = buffer[2*i+1] - 127;
	    }
	  free(buffer); free(zbuffer);
#else
	  printf("ERROR: This code cannot be decompressed with this program!\n");
	  exit(1);
#endif
	}
      else
	for(i=0;i<rangenum;i++)
	  {
	    r[i].umean = unpack(8,inp) - 127;
	    r[i].vmean = unpack(8,inp) - 127;
	  }
    }
  
  /* 4. Read in partition information */

  read_part(inp);
  fclose(inp);
  width *= basesize; height *= basesize;
  return 1;
}

void write_data(char *fname)
{
  FILE *out;
  int d,x,y;

  out = fopen(fname,"wb");
  if(!out) { printf("Error writing file %s!\n",fname); return; }
  fprintf(out,"P%d\n%d %d\n255\n",depth==3?6:5,width,height);
  for(y=0;y<height;y++)
    for(x=0;x<width;x++)
      for(d=0;d<depth;d++)
	fprintf(out,"%c",bound(image[d][y][x]));
  fclose(out);
}

void init_decoder()
{
  int i,j;

  image = (pixel ***) malloc(depth*sizeof(pixel **));
  image[0] = (pixel **) malloc(depth*height*sizeof(pixel *));
  image[0][0] = (pixel *) calloc(depth*height*width,sizeof(pixel));
  image2 = (pixel ***) malloc(depth*sizeof(pixel **));
  image2[0] = (pixel **) malloc(depth*height*sizeof(pixel *));
  image2[0][0] = (pixel *) calloc(depth*height*width,sizeof(pixel));
  for(i=0;i<depth;i++)
    {
      if(i)
	{
	  image[i] = image[i-1]+height;
	  image[i][0] = image[i-1][0]+height*width;
	  image2[i] = image2[i-1]+height;
	  image2[i][0] = image2[i-1][0]+height*width;
	}
      for(j=1;j<height;j++) 
	{
	  image[i][j] = image[i][j-1]+width;
	  image2[i][j] = image2[i][j-1]+width;
	}
    }
}

void clean_up()
{
  free(image[0][0]); free(image[0]); free(image);
  free(image2[0][0]); free(image2[0]); free(image2);
  free(r);
}

pixel subsample(pixel **img, int x, int y)
{
  return (img[y][x] + img[y+1][x] + img[y][x+1] + img[y+1][x+1]) / 4;
}

void add_partition(pixel ***image, rangeT *r,
		  int width, int height, int basesize, int rangenum)
{
  int j,k,l,x,y;
  int rx,ry,right,down;

  for(x=0;x<width;x++)
    for(y=0;y<height;y++)
      for(j=0;j<depth;j++)
	image[j][y][x] += (255-image[j][y][x])/8;

  width /= basesize;

  for(j=0;j<rangenum;j++)
    {
      for(k=0;k<r[j].size;k++)
	{
	  right = down = 1;
	  rx = basesize*(r[j].pos[k]%width);
	  ry = basesize*(r[j].pos[k]/width);
	  for(l=0;l<r[j].size;l++)
	    {
	      x = basesize*(r[j].pos[l]%width);
	      y = basesize*(r[j].pos[l]/width);
	      if(ry == y && (rx+basesize)%(basesize*width) == x) right = 0;
	      if(rx == x && (ry+basesize)%height == y) down = 0;
	    }
	  if(right)
	    for(y=0;y<basesize;y++)
	      for(l=0;l<depth;l++)
		image[l][y+ry][rx+basesize-1] = 0;
	  if(down)
	    for(x=0;x<basesize;x++)
	      for(l=0;l<depth;l++)
		image[l][ry+basesize-1][rx+x] = 0;
	}
    }

  width *= basesize;
}

void insert_color(pixel ***image, rangeT *r,
		  int width, int height, int basesize, int rangenum)
{
  int j,k,x,y;
  int rx,ry;

  width /= basesize;

  for(j=0;j<rangenum;j++)
    {
      /* copy transformed range */
      for(k=0;k<r[j].size;k++)
	{
	  rx = basesize*(r[j].pos[k]%width);
	  ry = basesize*(r[j].pos[k]/width);
	  for(x=0;x<basesize;x++)
	    for(y=0;y<basesize;y++)
	      {
		image[1][y+ry][x+rx] = r[j].umean;
		image[2][y+ry][x+rx] = r[j].vmean;
	      }
	}
    }

  width *= basesize;

  for(y=0;y<height;y++)
    for(x=0;x<width;x++)
      {
	pixel cy,cu,cv;

	cy = image[0][y][x];
	cu = image[1][y][x];
	cv = image[2][y][x];
	image[0][y][x] = cy              + 1.140 * cv;
	image[1][y][x] = cy - 0.394 * cu - 0.581 * cv;
	image[2][y][x] = cy + 2.028 * cu + 0.002 * cv;
      }
}

void decode_image(pixel ***image, pixel ***image2, rangeT *r,
		  int width, int height, int basesize, int rangenum, int iterations)
{
  int i,j,k,x,y,xx,yy,tmp;
  int rx,ry,dx,dy,neg,ddx,ddy,rot,domain;
  double s,o,val;
  pixel ***imtmp;

  width /= basesize; height /= basesize;

  if(iterations%2) { imtmp = image; image = image2; image2 = imtmp; }

  for(i=0;i<iterations;i++)
    {
      for(j=0;j<rangenum;j++)
	{
	  /* compute mean */
	  rot = r[j].rot;
	  domain = r[j].domain;
	  s = r[j].scaling;
	  o = r[j].offset;
	  neg = rot&4;
	  dx = 1 - (rot&2);
	  dy = 1 - ((rot&1)<<1);
	  
	  /* copy transformed range */
	  for(k=0;k<r[j].size;k++)
	    {
	      if(neg)
		{
		  ddx = ((width>>1)*height + domain%(width>>1) + 
			 dy*(r[j].pos[k]/width-r[j].pos[0]/width))%(width>>1);
		  ddy = (width*(height>>1) + domain/(width>>1) +
			 dx*(r[j].pos[k]%width-r[j].pos[0]%width))%(height>>1);
		}
	      else
		{
		  ddx = ((width>>1)*height + domain%(width>>1) +
			 dx*(r[j].pos[k]%width-r[j].pos[0]%width))%(width>>1);
		  ddy = (width*(height>>1) + domain/(width>>1) +
			 dy*(r[j].pos[k]/width-r[j].pos[0]/width))%(height>>1);
		}
	      ddx *= (basesize<<1); ddy *= (basesize<<1);
	      rx = basesize*(r[j].pos[k]%width);
	      ry = basesize*(r[j].pos[k]/width);
	      for(x=0;x<basesize;x++)
		for(y=0;y<basesize;y++)
		  {
		    xx = (dx == 1)?x:basesize-x-1;
		    yy = (dy == 1)?y:basesize-y-1;
		    if(neg) { tmp = xx; xx = yy; yy = tmp; }
		    val = s*subsample(image[0],ddx+2*xx,ddy+2*yy) + o;
		    if(val < 0.0) val = 0.0;
		    if(val > 255.0) val = 255.0;
		    image2[0][y+ry][x+rx] = rint(val);
		  }
	    }
	}
      imtmp = image; image = image2; image2 = imtmp;
    }
}

void parse_args(int argc, char **argv)
{
  int c;
  
  while ((c = getopt(argc,argv,"f:o:i:phv")) != EOF)
    {
      switch(c)
        {
        case 'f':
	  input = optarg; break;
        case 'o':
	  output = optarg; break;
        case 'i':
	  iterations = atoi(optarg); break;
	case 'p':
	  showpart = 1; break;
	case 'h':
	  printf("Usage: %s [-f <input>] [-o <output>] [-i <iterations] [-p] [-v]\n",argv[0]);
	  exit(0);
	  break;
	case 'v':
	  printf(VERSION); exit(0);
	  break;
        }
    } 
}

int main(int argc, char **argv)
{
  parse_args(argc,argv);
  if(!read_data(input)) return 1;
  init_decoder();
  decode_image(image,image2,r,width,height,basesize,rangenum,iterations);
  if(depth == 3) insert_color(image,r,width,height,basesize,rangenum);
  if(showpart) add_partition(image,r,width,height,basesize,rangenum);
  write_data(output);
  clean_up();
  return 0;
}
