/************************************************************************
    frap/defrap - fractal image compressor/decompressor
    Copyright (C) 1997-2000 Matthias Ruhl

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    The author can be contacted at:
      MIT Laboratory for Computer Science
      200 Technology Square
      Cambridge, MA 02139
      USA
    or via e-mail: ruhl@theory.lcs.mit.edu
*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#ifdef GETOPT
#include <getopt.h>
#endif
#include "frap.h"
#include "version.h"

static int rangemin = 1000;
static char *codefile = "frap.out";

#define max(a,b) ((a)>(b)?(a):(b))

void parse_args(int argc, char **argv)
{
  int c;
  
  while ((c = getopt(argc,argv,"b:c:f:s:r:o:e:hv")) != EOF)
    {
      switch(c)
        {
	case 'b':
	  basesize = atoi(optarg); break;
	case 's':
	  s_bits = atoi(optarg); break;
	case 'o':
	  o_bits = atoi(optarg); break;
	case 'e':
	  epsilon = atof(optarg); break;
        case 'f': 
	  if(image != NULL)
	    { fprintf(stderr,"Multiple filenames!\n"); continue; }
	  image = image_read(optarg);
	  break;
	case 'c':
	  codefile = optarg; break;
	case 'r':
	  rangemin = atoi(optarg); break;
	case 'h':
	  printf("Usage: %s <options>\n"\
		 "Options are: -f # (input imagefile)\n"\
		 "             -c # (codefilename)\n"\
		 "             -b # (atomic block size)\n"\
		 "             -r # (no. of ranges in code)\n"\
		 "             -s # (no. of bits for scaling factor)\n"\
		 "             -o # (no. of bits for offset parameter)\n"\
		 "             -e # (epsilon for nn-search)\n"\
		 "             -v   (output version of program)\n",argv[0]);
	  exit(0);
	  break;
	case 'v':
	  printf(VERSION); exit(0); break;
        }
    } 

  if(image == NULL)
    {
      printf("Input filename missing!\n"); exit(1);
    }
}

int main(int argc, char **argv)
{
  parse_args(argc,argv);
  init();
  while(rangenum>max(rangemin,50)) merge_best();
  output(codefile);
  image_free(image);
  free_all();
  return 0;
}
