/************************************************************************
    frap/defrap - fractal image compressor/decompressor
    Copyright (C) 1997-2000 Matthias Ruhl

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    The author can be contacted at:
      MIT Laboratory for Computer Science
      200 Technology Square
      Cambridge, MA 02139
      USA
    or via e-mail: ruhl@theory.lcs.mit.edu
*************************************************************************/

#ifndef FRAP_H
#define FRAP_H

typedef double pixel;

typedef struct
{
  int width,height,depth;
  char *filename;
  pixel ***img;
} imageT;

typedef struct
{
  /* transformation parameters */
  int domain,rot;
  int qoffset,qscaling;
  /* auxiliary variables */
  int precomp;
  double dsum,dsum2,rsum,rsum2,rdsum;
  double err;
} codeT;

typedef struct
{
  int size,*pos;
  int codenum;
  codeT *code;
  int timestamp,flag;
  double umean,vmean; /* for color images */
} rangeT;

typedef struct
{
  rangeT **r1,**r2,*join;
  int timestamp;
  double errcost,partcost;
} heapNodeT;

extern double epsilon;
extern int s_bits,o_bits;
extern imageT *image;

extern int basesize;
extern int codesize;
extern int width,height,depth,width2,height2;

/* pixel vectors of baseblocks */
extern int baseblocknum;
extern pixel **baseblock;
extern double *bsum,*bvar;

/* pixel vectors of shrunken domain blocks */
extern int codebooksize;
extern pixel **codebook;
extern double *csum,*cvar;

/* ranges in this partition */
extern int rangenum;
extern rangeT *range,**rptr;
  
/* rotation schemes */
extern int **rot;

/* Range-Domain Comparison: compare.c */

void rd_compare(rangeT *, int);

/* Output code file: encode.c */

void output(char *);

/* Partition functions: partition.c */

int gettime();
rangeT *merge_ranges(rangeT *, rangeT *);
void merge_best();
void init();
void free_all();

/* Image functions: image.c */

imageT *image_read(const char *);
void image_free(imageT *);

/* Heap functions: heap.c */

heapNodeT *heap_pop();
void heap_push(rangeT **, rangeT **);
void heap_init(int);
void heap_free();

#endif
