/************************************************************************
    frap/defrap - fractal image compressor/decompressor
    Copyright (C) 1997-2000 Matthias Ruhl

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    The author can be contacted at:
      MIT Laboratory for Computer Science
      200 Technology Square
      Cambridge, MA 02139
      USA
    or via e-mail: ruhl@theory.lcs.mit.edu
*************************************************************************/

#include <stdlib.h>
#include "frap.h"

static int heapmax,heapsize;
static heapNodeT *h;

static double costf(heapNodeT h)
{
  return h.errcost;
}

static double errcost(double join, double old1, double old2)
{
  return join-old1-old2;
}

static void recompute()
{
  free(h[1].join->code); 	  
  free(h[1].join->pos);
  free(h[1].join);
  h[1].join = merge_ranges(*(h[1].r1),*(h[1].r2));
  h[1].timestamp = gettime();
  h[1].errcost = errcost(h[1].join->code[0].err,
			 (*(h[1].r1))->code[0].err,(*(h[1].r2))->code[0].err);
}

static void siftup(int pos)
{
  heapNodeT tmp = h[pos];
  double cost = costf(tmp);

  while(pos > 1 && costf(h[pos>>1]) > cost)
    {
      h[pos] = h[pos>>1];
      pos >>= 1;
    }
  h[pos] = tmp;
}

static void siftdown(int pos)
{
  heapNodeT tmp = h[pos];

  pos <<= 1;
  while(pos <= heapsize)
    {
      if(pos < heapsize && costf(h[pos+1]) < costf(h[pos])) pos++;
      if(costf(h[pos]) >= costf(tmp)) break;
      h[pos>>1] = h[pos];
      pos <<= 1;
    }
  h[pos>>1] = tmp;
}

static void deletebest()
{
  if(heapsize == 0) return;
  h[1] = h[heapsize--];
  siftdown(1);
}

heapNodeT *heap_pop()
{
  while(heapsize > 0)
    {
      if(*(h[1].r1) == *(h[1].r2)) { deletebest(); continue; }
      if(h[1].timestamp < (*(h[1].r1))->timestamp ||
	 h[1].timestamp < (*(h[1].r2))->timestamp)
	{
	  recompute(); siftdown(1); continue;
	}
      h[0] = h[1];
      deletebest();
      return h;
    }
  return NULL;
}

void heap_push(rangeT **r1, rangeT **r2)
{
  heapsize++;
  h[heapsize].r1 = r1;
  h[heapsize].r2 = r2;
  h[heapsize].timestamp = gettime();
  h[heapsize].join = merge_ranges(*r1,*r2);
  h[heapsize].errcost = errcost(h[heapsize].join->code[0].err,
				(*r1)->code[0].err,(*r2)->code[0].err);
  siftup(heapsize);
}

void heap_init(int max)
{
  heapmax = max;
  heapsize = 0;
  h = (heapNodeT *) malloc((heapmax+1)*sizeof(heapNodeT));
}

