/************************************************************************
    frap/defrap - fractal image compressor/decompressor
    Copyright (C) 1997-2000 Matthias Ruhl

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    The author can be contacted at:
      MIT Laboratory for Computer Science
      200 Technology Square
      Cambridge, MA 02139
      USA
    or via e-mail: ruhl@theory.lcs.mit.edu
*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "frap.h"

static int error(const char *errmsg)
{
  fprintf(stderr,"%s\n",errmsg);
  return 0;
}

static int readint(FILE *inp)
{
  int c,n;

  while(1)
    {
      c = getc(inp);
      if(isdigit(c)) { ungetc(c,inp); break; }
      if(c == '#')
	while(c >= ' ') c = getc(inp);
      if(c == EOF) return -1;
    }
  fscanf(inp,"%d",&n);
  return n;
}

imageT *image_read(const char *fname)
{
  /* read image (PPM-format) from file */

  FILE *inp;
  unsigned char *buffer;
  int c,c1,i,j,k,l;
  imageT *image;
  
  inp = fopen(fname,"rb");
  if(!inp) { error("ppm_read: File not found!"); return NULL; }
  c = getc(inp); c1 = getc(inp);
  if(c != 'P' || (c1 != '5' && c1 != '6'))
    {
      fclose(inp);
      error("pgm_read: File not in PPM format!");
      return NULL;
    }
  /* read image parameters */
  image = (imageT *) malloc(sizeof(imageT));
  image->width = readint(inp);
  image->height = readint(inp);
  image->depth = (c1 == '6') ? 3 : 1;
  readint(inp);
  /* eat remaining whitespace */
  while((c = getc(inp)) >= ' ');

  buffer = (unsigned char *) malloc(image->depth*image->height*image->width);
  fread(buffer,image->depth*image->width*image->height,1,inp);
  fclose(inp);

  image->img = (pixel ***) malloc(image->depth*sizeof(pixel **));
  image->img[0] = (pixel **) malloc(image->depth*image->height*sizeof(pixel *));
  image->img[0][0] = (pixel *) malloc(image->depth*image->height*image->width*sizeof(pixel));
  for(i=0;i<image->depth;i++)
    {
      if(i)
	{
	  image->img[i] = image->img[i-1]+image->height;
	  image->img[i][0] = image->img[i-1][0]+image->height*image->width;
	}
      for(j=1;j<image->height;j++)
	image->img[i][j] = image->img[i][j-1]+image->width;
    }
  for(l=j=0;j<image->height;j++)
    for(k=0;k<image->width;k++)
      {
	for(i=0;i<image->depth;i++,l++)
	  image->img[i][j][k] = (pixel) buffer[l];
	if(image->depth == 3)
	  {
	    double r,g,b;
	    r = image->img[0][j][k];
	    g = image->img[1][j][k];
	    b = image->img[2][j][k];
	    image->img[0][j][k] =  0.299*r + 0.587*g + 0.114*b;
	    image->img[1][j][k] = -0.148*r - 0.289*g + 0.437*b;
	    image->img[2][j][k] =  0.615*r - 0.515*g - 0.100*b;
	  }
      }
  free(buffer);
  return image;
}

void image_free(imageT *image)
{
  if(image) { free(image->img[0][0]); free(image->img[0]); free(image->img); }
}
